unit UFormMain;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Menus, ComCtrls, ExtCtrls, ActnList, Math, dateutils, syncobjs, UProject,
  UConfig;

type

  TSectorState = (bsNone, bsOk, bsDamaged);

  { TBlockMap }

  TBlockMap = class
  private
    FBlockSize: TPoint;
    FDrawSize: TPoint;
    FSectorPerBlock: Integer;
    FSectors: array of TSectorState;
    FBlocks: array of TSectorState;
    function GetBlockCount: Integer;
    function GetSector(Index: Integer): TSectorState;
    function GetSectorCount: Integer;
    procedure SetBlockCount(AValue: Integer);
    procedure SetBlockSize(AValue: TPoint);
    procedure SetDrawSize(AValue: TPoint);
    procedure SetSector(Index: Integer; AValue: TSectorState);
    procedure SetSectorCount(AValue: Integer);
    procedure SetSectorPerBlock(AValue: Integer);
    procedure UpdateBlockSize;
    procedure UpdateBlocks;
  public
    ItemsCount: TPoint;
    function CombineSectors(From, Count: Integer): TSectorState;
    procedure Draw(Canvas: TCanvas);
    procedure Clear;
    constructor Create;
    property BlockSize: TPoint read FBlockSize write SetBlockSize;
    property DrawSize: TPoint read FDrawSize write SetDrawSize;
    property SectorPerBlock: Integer read FSectorPerBlock write SetSectorPerBlock;
    property SectorCount: Integer read GetSectorCount write SetSectorCount;
    property BlockCount: Integer read GetBlockCount;
    property Sectors[Index: Integer]: TSectorState read GetSector write SetSector;
  end;

  TRunMode = (rmRead, rmWrite);

  { TDriveScan }

  TDriveScan = class
  private
    FOnChange: TNotifyEvent;
    FOnTerminate: TNotifyEvent;
    procedure DoChange;
    procedure DoTerminate;
  public
    Lock: TCriticalSection;
    BlockMap: TBlockMap;
    SectorSize: Integer;
    SectorCurrent: Integer;
    TimeStart: TDateTime;
    TimeEnd: TDateTime;
    Terminated: Boolean;
    DamagedBlockCount: Integer;
    Drive: string;
    Mode: TRunMode;
    WritePattern: Byte;
    function GetElapsedTime: TDateTime;
    procedure Run;
    constructor Create;
    destructor Destroy; override;
    property OnTerminate: TNotifyEvent read FOnTerminate write FOnTerminate;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TScanThread }

  TScanThread = class(TThread)
    Form: TForm;
    procedure Execute; override;
  end;

  { TFormMain }

  TFormMain = class(TForm)
    ABrowseDisk: TAction;
    AScanOptions: TAction;
    AScanStart: TAction;
    AScanStop: TAction;
    ActionList1: TActionList;
    Button1: TButton;
    ButtonScan: TButton;
    ButtonScan1: TButton;
    ButtonScan2: TButton;
    ComboBoxRunMode: TComboBox;
    EditPattern: TEdit;
    EditDrive: TEdit;
    Image1: TImage;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    LabelElapsedTime: TLabel;
    LabelBlockSize: TLabel;
    LabelBlockCount: TLabel;
    LabelBlockCurrent: TLabel;
    LabelBlockDamaged: TLabel;
    LabelEstimatedTime: TLabel;
    LabelIOSpeed: TLabel;
    LabelSectorPerBlock: TLabel;
    ListView1: TListView;
    MainMenu1: TMainMenu;
    OpenDialog1: TOpenDialog;
    Timer1: TTimer;
    procedure ABrowseDiskExecute(Sender: TObject);
    procedure AScanOptionsExecute(Sender: TObject);
    procedure AScanStartExecute(Sender: TObject);
    procedure AScanStopExecute(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Image1Resize(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    LastBlockPos: Integer;
    DriveScan: TDriveScan;
    ScanThread: TScanThread;
    RedrawPending: Boolean;
    procedure DriveScanChange(Sender: TObject);
    procedure DriveScanTerminate(Sender: TObject);
    procedure Redraw(Delayed: Boolean = False);
    procedure DoDraw;
    procedure UpdateInterface;
  public
    Project: TProject;
    procedure Detect;
    procedure ShowBuffer(Buffer: array of Byte);
  end;

var
  FormMain: TFormMain;

implementation

{$R *.lfm}

uses
  UFormProject;

{ TScanThread }

procedure TScanThread.Execute;
begin
  TFormMain(Form).DriveScan.Run;
end;

{ TBlockMap }

function TBlockMap.GetSectorCount: Integer;
begin
  Result := Length(FSectors);
end;

procedure TBlockMap.SetBlockCount(AValue: Integer);
begin
  SetLength(FBlocks, AValue);
end;

procedure TBlockMap.SetBlockSize(AValue: TPoint);
begin
  if (FBlockSize.X = AValue.X) and (FBlockSize.Y = AValue.Y) then Exit;
  if (AValue.X = 0) or (AValue.Y = 0) then
    raise Exception.Create('BlockSize can''t be set to 0');
  FBlockSize := AValue;
end;

procedure TBlockMap.SetDrawSize(AValue: TPoint);
begin
  if (FDrawSize.X = AValue.X) and (FDrawSize.Y = AValue.Y) then Exit;
  FDrawSize := AValue;
  UpdateBlockSize;
end;

function TBlockMap.GetSector(Index: Integer): TSectorState;
begin
  Result := FSectors[Index];
end;

function TBlockMap.GetBlockCount: Integer;
begin
  Result := Length(FBlocks);
end;

procedure TBlockMap.SetSector(Index: Integer; AValue: TSectorState);
var
  BlockIndex: Integer;
begin
  FSectors[Index] := AValue;

  // Update block
  BlockIndex := Index div FSectorPerBlock;
  if (AValue = bsDamaged) then FBlocks[BlockIndex] := bsDamaged
  else if (AValue = bsOk) and (FBlocks[BlockIndex] = bsNone) then
    FBlocks[BlockIndex] := bsOk;
end;

procedure TBlockMap.SetSectorCount(AValue: Integer);
begin
  if Length(FSectors) = AValue then Exit;
  SetLength(FSectors, AValue);
  UpdateBlockSize;
end;

procedure TBlockMap.SetSectorPerBlock(AValue: Integer);
begin
  if FSectorPerBlock = AValue then Exit;
  FSectorPerBlock := AValue;
  UpdateBlockSize;
end;

procedure TBlockMap.UpdateBlockSize;
begin
  ItemsCount := Point(Trunc(FDrawSize.X / BlockSize.X), Trunc(FDrawSize.Y / BlockSize.Y));
  FSectorPerBlock := Ceil(SectorCount / (ItemsCount.X * ItemsCount.Y));
  SetBlockCount(Ceil(SectorCount / FSectorPerBlock));
  UpdateBlocks;
end;

procedure TBlockMap.UpdateBlocks;
var
  I: Integer;
begin
  for I := 0 to BlockCount - 1 do
    FBlocks[I] := CombineSectors(I * FSectorPerBlock, FSectorPerBlock);
end;

function TBlockMap.CombineSectors(From, Count: Integer): TSectorState;
var
  I: Integer;
  UpTo: Integer;
begin
  Result := bsNone;
  UpTo := From + Count - 1;
  if UpTo >= SectorCount then Upto := SectorCount - 1;
  for I := From to UpTo do begin

    if (Result = bsNone) and (Sectors[I] = bsOk) then Result := bsOk
    else if (Result = bsOk) and (Sectors[I] = bsDamaged) then Result := bsDamaged;
  end;
end;

procedure TBlockMap.Draw(Canvas: TCanvas);
var
  I: Integer;
  Rect: TRect;
  BlockState: TSectorState;
begin
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Color := clBlack;
  // Clean background
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := clBlack;
  Canvas.FillRect(0, 0, Canvas.Width, Canvas.Height);

  DrawSize := Point(Canvas.Width, Canvas.Height);
  if SectorPerBlock >= 1 then begin
    for I := 0 to BlockCount - 1 do begin
      Rect := Bounds((I mod ItemsCount.X) * BlockSize.X,
        (I div ItemsCount.X) * BlockSize.Y, BlockSize.X - 1, BlockSize.Y - 1);
      BlockState := FBlocks[I];
      if BlockState = bsOk then Canvas.Brush.Color := clGreen
        else if BlockState = bsDamaged then Canvas.Brush.Color := clRed
        else Canvas.Brush.Color := clSilver;
      Canvas.FillRect(Rect);
      //Canvas.TextOut(Rect.Left, Rect.Top, IntToStr(I));
    end;
  end;
end;

procedure TBlockMap.Clear;
var
  I: Integer;
begin
  for I := 0 to SectorCount - 1 do
    FSectors[I] := bsNone;
end;

constructor TBlockMap.Create;
begin
  FBlockSize := Point(12, 12);
  SectorPerBlock := 1;
end;

{ TDriveScan }

procedure TDriveScan.DoChange;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDriveScan.DoTerminate;
begin
  if Assigned(FOnTerminate) then FOnTerminate(Self);
end;

function TDriveScan.GetElapsedTime: TDateTime;
begin
  if TimeEnd <> 0 then Result := TimeEnd - TimeStart
    else Result := Now - TimeStart;
end;

procedure TDriveScan.Run;
var
  F: TFileStream;
  RealSize: Integer;
  Buffer: array of Byte;
  I: Integer;
begin
  try
  Lock.Acquire;
  TimeStart := Now;
  Terminated := False;
  DamagedBlockCount := 0;
  if Mode = rmRead then F := TFileStream.Create(Drive, fmOpenRead)
    else if Mode = rmWrite then F := TFileStream.Create(Drive, fmOpenReadWrite);
  BlockMap.SectorCount := F.Size div SectorSize;
  BlockMap.Clear;
  SetLength(Buffer, SectorSize);
  if Mode = rmWrite then
    FillChar(Buffer[0], Length(Buffer), WritePattern);
  for I := 0 to BlockMap.SectorCount - 1 do begin
    SectorCurrent := I;
    try
      Lock.Release;
      DoChange;

      if ConfigTest then begin
        if Random < 0.000001 then RealSize := 0
          else RealSize := SectorSize;
        //Sleep(1);
      end else begin
        F.Position := I * SectorSize;
        if Mode = rmRead then RealSize := F.Read(Buffer[0], SectorSize)
          else if Mode = rmWrite then RealSize := F.Write(Buffer[0], SectorSize)
          else raise Exception.Create('Unknwon run mode');
      end;
    finally
      Lock.Acquire;
    end;
    if RealSize <> SectorSize then begin
      BlockMap.Sectors[I] := bsDamaged;
      Inc(DamagedBlockCount);
    end else BlockMap.Sectors[I] := bsOk;
    if Terminated then Break;
  end;
  F.Free;

  finally
    Lock.Release;
  end;
  TimeEnd := Now;
  DoChange;
  Terminated := True;
  DoTerminate;
end;

constructor TDriveScan.Create;
begin
  Lock := TCriticalSection.Create;
  BlockMap := TBlockMap.Create;
  SectorSize := 4096;
  Terminated := True;
end;

destructor TDriveScan.Destroy;
begin
  FreeAndNil(BlockMap);
  FreeAndNil(Lock);
  inherited Destroy;
end;

{ TFormMain }

procedure TFormMain.FormShow(Sender: TObject);
begin
  Redraw;
  UpdateInterface;
end;

procedure TFormMain.Image1Resize(Sender: TObject);
begin
  Redraw;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin
  DriveScan := TDriveScan.Create;
  DriveScan.OnChange := DriveScanChange;
  DriveScan.OnTerminate := DriveScanTerminate;
  ScanThread := TScanThread.Create(True);
  ScanThread.Form := Self;
  Project := TProject.Create;
  EditDrive.Text := '/dev/sdb';
end;

procedure TFormMain.AScanStopExecute(Sender: TObject);
begin
  DriveScan.Terminated := True;
  ScanThread.Terminate;
  ScanThread.WaitFor;
  UpdateInterface;
end;

procedure TFormMain.AScanStartExecute(Sender: TObject);
begin
  LastBlockPos := 0;
  DriveScan.Drive := EditDrive.Text;
  DriveScan.SectorSize := Project.SectorSize;
  DriveScan.Terminated := False;
  DriveScan.Mode := TRunMode(ComboBoxRunMode.ItemIndex);
  DriveScan.WritePattern := StrToInt(EditPattern.Text);
  ScanThread.Start;
  UpdateInterface;
end;

procedure TFormMain.AScanOptionsExecute(Sender: TObject);
begin
  FormProject.Load(Project);
  if FormProject.ShowModal = mrOk then begin
    FormProject.Save(Project);
  end;
end;

procedure TFormMain.ABrowseDiskExecute(Sender: TObject);
begin
  OpenDialog1.Title := 'Select drive for scan';
  OpenDialog1.FileName := EditDrive.Text;
  if OpenDialog1.Execute then
    EditDrive.Text := OpenDialog1.FileName;
end;

procedure TFormMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  DriveScan.Terminated := True;
end;

procedure TFormMain.FormDestroy(Sender: TObject);
begin
  FreeAndNil(Project);
  FreeAndNil(ScanThread);
  FreeAndNil(DriveScan);
end;

procedure TFormMain.Redraw(Delayed: Boolean = False);
begin
  RedrawPending := True;
  if not Delayed then DoDraw;
end;

procedure TFormMain.DoDraw;
begin
  if RedrawPending then
  with Image1 do begin
    if (Width <> Picture.Bitmap.Width) or (Height <> Picture.Bitmap.Height) then
      Picture.Bitmap.SetSize(Width, Height);
    try
      DriveScan.Lock.Acquire;
      try
        Picture.Bitmap.BeginUpdate(True);
        DriveScan.BlockMap.Draw(Picture.Bitmap.Canvas);
      finally
        Picture.Bitmap.EndUpdate;
      end;
      LabelSectorPerBlock.Caption := IntToStr(DriveScan.BlockMap.SectorPerBlock);
      LabelBlockSize.Caption := IntToStr(DriveScan.SectorSize) + ' bytes';
      LabelBlockCount.Caption := IntToStr(DriveScan.BlockMap.SectorCount);
      LabelBlockCurrent.Caption := IntToStr(DriveScan.SectorCurrent);
      LabelBlockDamaged.Caption := IntToStr(DriveScan.DamagedBlockCount);
    finally
      DriveScan.Lock.Release;
    end;
    RedrawPending := False;
  end;
  try
    DriveScan.Lock.Acquire;
    if not DriveScan.Terminated then begin
      LabelElapsedTime.Caption := TimeToStr(DriveScan.GetElapsedTime);
      LabelEstimatedTime.Caption := TimeToStr((Now - DriveScan.TimeStart) /
        DriveScan.SectorCurrent * (DriveScan.BlockMap.SectorCount - DriveScan.SectorCurrent));
      LabelIOSpeed.Caption := IntToStr(Trunc((DriveScan.SectorCurrent - LastBlockPos) *
        DriveScan.SectorSize / (1024 * 1024) / (Timer1.Interval / 1000))) + ' MB/s';
      LastBlockPos := DriveScan.SectorCurrent;
    end else begin
      LabelElapsedTime.Caption := '';
      LabelEstimatedTime.Caption := '';
      LabelIOSpeed.Caption := '';
    end;
  finally
    DriveScan.Lock.Release;
  end;
end;

procedure TFormMain.UpdateInterface;
begin
  AScanStart.Enabled := DriveScan.Terminated = True;
  AScanStop.Enabled := DriveScan.Terminated = False;
end;

procedure TFormMain.Timer1Timer(Sender: TObject);
begin
  Redraw;
end;

procedure TFormMain.DriveScanChange(Sender: TObject);
begin
  RedrawPending := True;
end;

procedure TFormMain.DriveScanTerminate(Sender: TObject);
begin
  UpdateInterface;
end;

procedure TFormMain.Detect;
begin

end;

procedure TFormMain.ShowBuffer(Buffer: array of Byte);
const
  BytePerRow = 16;
var
  Row: string;
  I: Integer;
  Item: TListItem;
begin
  ListView1.Clear;
  Row := '';
  for I := 0 to Length(Buffer) do begin
    Row := Row + IntToHex(Buffer[I], 2) + ' ';
    if ((I mod BytePerRow) = (BytePerRow - 1)) or (I = (Length(Buffer) - 1)) then begin
      Item := ListView1.Items.Add;
      Item.Caption := '';
      Item.SubItems.Add(Row);
      Row := '';
    end;
  end;
end;

end.

